# THIS FILE IS GENERATED FROM OKONOMIYAKI SETUP.PY
version = '0.14.0'
full_version = '0.14.0'
git_revision = '3e5c81393de2c5cbef7d236103c2e26a1f49d4d0'
is_released = True

if is_released:
    version_info = (0, 14, 0, 'final', 0)
else:
    version = full_version
    version_info = (0, 14, 0, 'dev', 639)
