# THIS FILE IS GENERATED FROM ENSTALLER SETUP.PY
version = '4.8.12'
full_version = '4.8.12'
git_revision = '5b5bcc3b4b563d0de42417c6a805e0dd1f74a29a'
is_released = True

if not is_released:
    version = full_version
