# THIS FILE IS GENERATED FROM ZIPFILE2 SETUP.PY
version = '0.0.11'
full_version = '0.0.11'
git_revision = '5958ad16d076e8c2eec06f55bca66e1dd9c5a1b4'
is_released = True

if not is_released:
    version = full_version
